class DaemonError(Exception):
    """
    Raised when the GoXLR Utility returns an 'Error' message.
    """

    pass


class MixerNotFoundError(Exception):
    """
    Raised when a given mixer does not exist.
    """

    pass


class MissingFeatureError(Exception):
    """
    Raised when a feature is not available on the GoXLR Mini.
    """

    pass



from ..types import LogLevel, PathType


class DaemonCommands:
    """
    Used to send commands to the GoXLR daemon.
    """

    async def __send_daemon(self, payload):
        payload = {"Daemon": payload}
        return await self.send(payload)

    async def open_ui(self):
        return await self.__send_daemon("OpenUi")

    async def activate(self):
        return await self.__send_daemon("Activate")

    async def stop_daemon(self):
        return await self.__send_daemon("StopDaemon")

    async def open_path(self, pathType: PathType):
        return await self.__send_daemon({"OpenPath": pathType.name})

    async def set_log_level(self, logLevel: LogLevel):
        return await self.__send_daemon({"SetLogLevel": logLevel.name})

    async def set_show_tray_icon(self, enabled: bool):
        return await self.__send_daemon({"SetShowTrayIcon": enabled})

    async def set_tts_enabled(self, enabled: bool):
        return await self.__send_daemon({"SetTTSEnabled": enabled})

    async def set_auto_start_enabled(self, enabled: bool):
        return await self.__send_daemon({"SetAutoStartEnabled": enabled})

    async def set_allow_network_access(self, enabled: bool):
        return await self.__send_daemon({"SetAllowNetworkAccess": enabled})

    async def recover_defaults(self, pathType: PathType):
        return await self.__send_daemon({"RecoverDefaults": pathType.name})



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional
from .enums import *

# --------------------------------------------------
# Config
# --------------------------------------------------


@dataclass
class HttpSettings:
    enabled: bool
    bind_address: str
    cors_enabled: bool
    port: int

    def __init__(self, http_settings: dict):
        self.enabled = http_settings.get("enabled")
        self.bind_address = http_settings.get("bind_address")
        self.cors_enabled = http_settings.get("cors_enabled")
        self.port = http_settings.get("port")


@dataclass
class Config:
    http_settings: HttpSettings
    daemon_version: str
    autostart_enabled: bool
    show_tray_icon: bool
    tts_enabled: bool
    allow_network_access: bool
    log_level: LogLevel

    def __init__(self, config: dict):
        self.http_settings = HttpSettings(config.get("http_settings"))
        self.daemon_version = config.get("daemon_version")
        self.autostart_enabled = config.get("autostart_enabled")
        self.show_tray_icon = config.get("show_tray_icon")
        self.tts_enabled = config.get("tts_enabled")
        self.allow_network_access = config.get("allow_network_access")
        self.log_level = LogLevel[config.get("log_level")]


# --------------------------------------------------
# Mixer - Hardware
# --------------------------------------------------


@dataclass
class MixerVersions:
    firmware: List[int]
    fpga_count: int
    dice: List[int]

    def __init__(self, mixer_version: dict):
        self.firmware = mixer_version.get("firmware")
        self.fpga_count = mixer_version.get("fpga_count")
        self.dice = mixer_version.get("dice")


@dataclass
class USBDevice:
    manufacturer_name: str
    product_name: str
    version: List[int]
    bus_number: int
    address: int
    identifier: str

    def __init__(self, usb_device: dict):
        self.manufacturer_name = usb_device.get("manufacturer_name")
        self.product_name = usb_device.get("product_name")
        self.version = usb_device.get("version")
        self.bus_number = usb_device.get("bus_number")
        self.address = usb_device.get("address")
        self.identifier = usb_device.get("identifier")


@dataclass
class HardwareInfo:
    versions: MixerVersions
    serial_number: str
    manufactured_date: datetime
    device_type: DeviceType
    usb_device: USBDevice

    def __init__(self, hardware_info: dict):
        self.versions = MixerVersions(hardware_info.get("versions"))
        self.serial_number = hardware_info.get("serial_number")
        self.manufactured_date = datetime.strptime(
            hardware_info.get("manufactured_date"), "%Y-%m-%d"
        )
        self.device_type = hardware_info.get("device_type")
        self.usb_device = USBDevice(hardware_info.get("usb_device"))


# --------------------------------------------------
# Mixer - Fader Status
# --------------------------------------------------


@dataclass
class Scribble:
    file_name: str
    bottom_text: str
    left_text: str
    inverted: bool

    def __init__(self, scribble: dict):
        self.file_name = scribble.get("file_name")
        self.bottom_text = scribble.get("bottom_text")
        self.left_text = scribble.get("left_text")
        self.inverted = scribble.get("inverted")


@dataclass
class FaderStatus:
    channel: Channel
    mute_type: MuteFunction
    scribble: Scribble
    mute_state: MuteState

    def __init__(self, fader_status: dict):
        self.channel = Channel[fader_status.get("channel")]
        self.mute_type = MuteFunction[fader_status.get("mute_type")]
        self.scribble = Scribble(fader_status.get("scribble"))
        self.mute_state = MuteState[fader_status.get("mute_state")]


# --------------------------------------------------
# Mixer - Fader Status
# --------------------------------------------------


@dataclass
class Equaliser:
    gain: Dict[EqFrequency, int]
    frequency: Dict[EqFrequency, float]

    def __init__(self, equaliser: dict):
        self.gain = {EqFrequency[k]: v for k, v in equaliser.get("gain").items()}
        self.frequency = {
            EqFrequency[k]: v for k, v in equaliser.get("frequency").items()
        }


@dataclass
class EqMini(Equaliser):
    gain: Dict[MiniEqFrequency, int]
    frequency: Dict[MiniEqFrequency, float]

    def __init__(self, eq_mini: dict):
        self.gain = {MiniEqFrequency[k]: v for k, v in eq_mini.get("gain").items()}
        self.frequency = {
            MiniEqFrequency[k]: v for k, v in eq_mini.get("frequency").items()
        }


@dataclass
class NoiseGate:
    threshold: int
    attack: int
    release: int
    enabled: bool
    attenuation: int

    def __init__(self, noise_gate: dict):
        self.threshold = noise_gate.get("threshold")
        self.attack = noise_gate.get("attack")
        self.release = noise_gate.get("release")
        self.enabled = noise_gate.get("enabled")
        self.attenuation = noise_gate.get("attenuation")


@dataclass
class Compressor:
    threshold: int
    ratio: int
    attack: int
    release: int
    makeup_gain: int

    def __init__(self, compressor: dict):
        self.threshold = compressor.get("threshold")
        self.ratio = compressor.get("ratio")
        self.attack = compressor.get("attack")
        self.release = compressor.get("release")
        self.makeup_gain = compressor.get("makeup_gain")


@dataclass
class MicStatus:
    mic_type: MicrophoneType
    mic_gains: Dict[MicrophoneType, int]
    equaliser: Equaliser
    equaliser_mini: EqMini
    noise_gate: NoiseGate
    compressor: Compressor

    def __init__(self, mic_status: dict):
        self.mic_type = MicrophoneType[mic_status.get("mic_type")]

        # example: {"Dynamic": 30, "Condenser": 40, "Jack": 30}
        self.mic_gains = {
            MicrophoneType[k]: v for k, v in mic_status.get("mic_gains").items()
        }

        self.equaliser = Equaliser(mic_status.get("equaliser"))
        self.equaliser_mini = EqMini(mic_status.get("equaliser_mini"))

        self.noise_gate = NoiseGate(mic_status.get("noise_gate"))
        self.compressor = Compressor(mic_status.get("compressor"))


# --------------------------------------------------
# Mixer - Levels
# --------------------------------------------------


@dataclass
class Submix:
    volume: int
    linked: bool
    ratio: float

    def __init__(self, submix: dict):
        self.volume = submix.get("volume")
        self.linked = submix.get("linked")
        self.ratio = submix.get("ratio")


@dataclass
class Submixes:
    inputs: Dict[SubMixChannel, Submix]
    outputs: Dict[OutputDevice, Mix]

    def __init__(self, submixes: dict):
        self.inputs = {
            SubMixChannel[k]: Submix(v) for k, v in submixes.get("inputs").items()
        }
        self.outputs = {
            OutputDevice[k]: Mix[v] for k, v in submixes.get("outputs").items()
        }


@dataclass
class Levels:
    submix_supported: bool
    output_monitor: OutputDevice
    volumes: Dict[Channel, int]
    submix: Optional[Submixes]
    bleep: int
    deess: int

    def __init__(self, levels: dict):
        self.submix_supported = levels.get("submix_supported")
        self.output_monitor = OutputDevice[levels.get("output_monitor")]
        self.volumes = {Channel[k]: v for k, v in levels.get("volumes").items()}
        if submix := levels.get("submix"):
            self.submix = {Channel[k]: v for k, v in submix.items()}
        else:
            self.submix = None
        self.bleep = levels.get("bleep")
        self.deess = levels.get("deess")


# --------------------------------------------------
# Mixer - Cough Button
# --------------------------------------------------


@dataclass
class CoughButton:
    is_toggle: bool
    mute_type: MuteFunction
    state: MuteState

    def __init__(self, cough_button: dict):
        self.is_toggle = cough_button.get("is_toggle")
        self.mute_type = MuteFunction[cough_button.get("mute_type")]
        self.mute_state = MuteState[cough_button.get("state")]


# --------------------------------------------------
# Mixer - Lighting
# --------------------------------------------------


@dataclass
class Animation:
    supported: bool
    mode: AnimationMode
    mod1: int
    mod2: int
    waterfall_direction: WaterfallDirection

    def __init__(self, animation: dict):
        self.supported = animation.get("supported")

        # if "None" set to AnimationMode.NONE else AnimationMode[animation.get("mode")]
        # use walrus operator instead of getting the value twice
        self.mode = (
            AnimationMode.NONE
            if (mode := animation.get("mode")) == "None"
            else AnimationMode[mode]
        )

        self.mod1 = animation.get("mod1")
        self.mod2 = animation.get("mod2")
        self.waterfall_direction = WaterfallDirection[
            animation.get("waterfall_direction")
        ]


@dataclass
class Colours:
    """
    Store up to three colours for a lighting style.

    Can be initialised with a dict or up to three strings.
    """

    colour_one: str
    colour_two: Optional[str]
    colour_three: Optional[str]

    def __init__(self, colour1: dict | str, colour2: str = None, colour3: str = None):
        if type(colour1) == str:
            self.colour_one = colour1
            self.colour_two = colour2
            self.colour_three = colour3
        else:
            self.colour_one = colour1.get("colour_one")
            self.colour_two = colour1.get("colour_two")
            self.colour_three = colour1.get("colour_three")


@dataclass
class FaderLighting:
    style: FaderDisplayStyle
    colours: Colours

    def __init__(self, lighting: dict):
        self.style = FaderDisplayStyle[lighting.get("style")]
        self.colours = Colours(lighting.get("colours"))


@dataclass
class ButtonLighting:
    off_style: ButtonColourOffStyle
    colours: Colours

    def __init__(self, lighting: dict):
        self.off_style = ButtonColourOffStyle[lighting.get("off_style")]
        self.colours = Colours(lighting.get("colours"))


@dataclass
class Lighting:
    animation: Animation
    faders: Dict[Fader, FaderLighting]
    buttons: Dict[Button, ButtonLighting]
    simple: Dict[SimpleColourTarget, Colours]
    sampler: Dict[SamplerColourTarget, ButtonLighting]
    encoders: Dict[Encoder, Colours]

    def __init__(self, lighting: dict):
        self.animation = Animation(lighting.get("animation"))
        self.faders = {
            Fader[k]: FaderLighting(v) for k, v in lighting.get("faders").items()
        }
        self.buttons = {
            Button[k]: ButtonLighting(v) for k, v in lighting.get("buttons").items()
        }
        self.simple = {
            SimpleColourTarget[k]: Colours(v) for k, v in lighting.get("simple").items()
        }
        self.sampler = {
            SamplerColourTarget[k]: ButtonLighting(v)
            for k, v in lighting.get("sampler").items()
        }
        self.encoders = {
            Encoder[k]: Colours(v) for k, v in lighting.get("encoders").items()
        }


# --------------------------------------------------
# Mixer - Effects
# --------------------------------------------------


@dataclass
class Reverb:
    style: ReverbStyle
    amount: int
    decay: int
    early_level: int
    tail_level: int
    pre_delay: int
    lo_colour: int
    hi_colour: int
    hi_factor: int
    diffuse: int
    mod_speed: int
    mod_depth: int

    def __init__(self, reverb: dict):
        self.style = ReverbStyle[reverb.get("style")]
        self.amount = reverb.get("amount")
        self.decay = reverb.get("decay")
        self.early_level = reverb.get("early_level")
        self.tail_level = reverb.get("tail_level")
        self.pre_delay = reverb.get("pre_delay")
        self.lo_colour = reverb.get("lo_colour")
        self.hi_colour = reverb.get("hi_colour")
        self.hi_factor = reverb.get("hi_factor")
        self.diffuse = reverb.get("diffuse")
        self.mod_speed = reverb.get("mod_speed")
        self.mod_depth = reverb.get("mod_depth")


@dataclass
class Echo:
    style: EchoStyle
    amount: int
    feedback: int
    tempo: int
    delay_left: int
    delay_right: int
    feedback_left: int
    feedback_right: int
    feedback_xfb_l_to_r: int
    feedback_xfb_r_to_l: int

    def __init__(self, echo: dict):
        self.style = EchoStyle[echo.get("style")]
        self.amount = echo.get("amount")
        self.feedback = echo.get("feedback")
        self.tempo = echo.get("tempo")
        self.delay_left = echo.get("delay_left")
        self.delay_right = echo.get("delay_right")
        self.feedback_left = echo.get("feedback_left")
        self.feedback_right = echo.get("feedback_right")
        self.feedback_xfb_l_to_r = echo.get("feedback_xfb_l_to_r")
        self.feedback_xfb_r_to_l = echo.get("feedback_xfb_r_to_l")


@dataclass
class Pitch:
    style: PitchStyle
    amount: int
    character: int

    def __init__(self, pitch: dict):
        self.style = PitchStyle[pitch.get("style")]
        self.amount = pitch.get("amount")
        self.character = pitch.get("character")


@dataclass
class Gender:
    style: GenderStyle
    amount: int

    def __init__(self, gender: dict):
        self.style = GenderStyle[gender.get("style")]
        self.amount = gender.get("amount")


@dataclass
class Megaphone:
    is_enabled: bool
    style: MegaphoneStyle
    amount: int
    post_gain: int

    def __init__(self, megaphone: dict):
        self.is_enabled = megaphone.get("is_enabled")
        self.style = MegaphoneStyle[megaphone.get("style")]
        self.amount = megaphone.get("amount")
        self.post_gain = megaphone.get("post_gain")


@dataclass
class Robot:
    is_enabled: bool
    style: RobotStyle
    low_gain: int
    low_freq: int
    low_width: int
    mid_gain: int
    mid_freq: int
    mid_width: int
    high_gain: int
    high_freq: int
    high_width: int
    waveform: int
    pulse_width: int
    threshold: int
    dry_mix: int

    def __init__(self, robot: dict):
        self.is_enabled = robot.get("is_enabled")
        self.style = RobotStyle[robot.get("style")]
        self.low_gain = robot.get("low_gain")
        self.low_freq = robot.get("low_freq")
        self.low_width = robot.get("low_width")
        self.mid_gain = robot.get("mid_gain")
        self.mid_freq = robot.get("mid_freq")
        self.mid_width = robot.get("mid_width")
        self.high_gain = robot.get("high_gain")
        self.high_freq = robot.get("high_freq")
        self.high_width = robot.get("high_width")
        self.waveform = robot.get("waveform")
        self.pulse_width = robot.get("pulse_width")
        self.threshold = robot.get("threshold")
        self.dry_mix = robot.get("dry_mix")


@dataclass
class HardTune:
    is_enabled: bool
    style: HardTuneStyle
    amount: int
    rate: int
    window: int
    source: HardTuneSource

    def __init__(self, hard_tune: dict):
        self.is_enabled = hard_tune.get("is_enabled")
        self.style = HardTuneStyle[hard_tune.get("style")]
        self.amount = hard_tune.get("amount")
        self.rate = hard_tune.get("rate")
        self.window = hard_tune.get("window")
        self.source = HardTuneSource[hard_tune.get("source")]


@dataclass
class CurrentEffects:
    reverb: Reverb
    echo: Echo
    pitch: Pitch
    gender: Gender
    megaphone: Megaphone
    robot: Robot
    hard_tune: HardTune

    def __init__(self, effects: dict):
        self.reverb = Reverb(effects.get("reverb"))
        self.echo = Echo(effects.get("echo"))
        self.pitch = Pitch(effects.get("pitch"))
        self.gender = Gender(effects.get("gender"))
        self.megaphone = Megaphone(effects.get("megaphone"))
        self.robot = Robot(effects.get("robot"))
        self.hard_tune = HardTune(effects.get("hard_tune"))


@dataclass
class Effects:
    is_enabled: bool
    active_preset: EffectBankPreset
    preset_names: Dict[EffectBankPreset, str]
    current: CurrentEffects

    def __init__(self, effects: dict):
        self.is_enabled = effects.get("is_enabled")
        self.active_preset = EffectBankPreset[effects.get("active_preset")]
        self.preset_names = effects.get("preset_names")
        self.current = CurrentEffects(effects.get("current"))


# -------------------------------------------------------
# Mixer - Sampler
# -------------------------------------------------------


@dataclass
class Sample:
    name: str
    start_pct: float
    stop_pct: float

    def __init__(self, sample: dict):
        self.name = sample.get("name")
        self.start_pct = sample.get("start_pct")
        self.stop_pct = sample.get("stop_pct")


@dataclass
class SampleMetadata:
    function: SamplePlaybackMode
    order: SamplePlayOrder
    samples: List[Sample]
    is_playing: bool
    is_recording: bool

    def __init__(self, sample_button: dict):
        self.function = SamplePlaybackMode[sample_button.get("function")]
        self.order = SamplePlayOrder[sample_button.get("order")]
        self.samples = [Sample(sample) for sample in sample_button.get("samples")]
        self.is_playing = sample_button.get("is_playing")
        self.is_recording = sample_button.get("is_recording")


@dataclass
class SamplerProcessState:
    progress: Optional[int]
    last_error: Optional[str]

    def __init__(self, process_state: dict):
        self.progress = process_state.get("progress")
        self.last_error = process_state.get("last_error")


@dataclass
class Sampler:
    processing_state: SamplerProcessState
    active_bank: SampleBank
    clear_active: bool
    record_buffer: int
    banks: Dict[SampleBank, Dict[SampleButton, SampleMetadata]]

    def __init__(self, sampler: dict):
        self.processing_state = SamplerProcessState(sampler.get("processing_state"))
        self.active_bank = SampleBank[sampler.get("active_bank")]
        self.clear_active = sampler.get("clear_active")
        self.record_buffer = sampler.get("record_buffer")
        self.banks = sampler.get("banks")


# -------------------------------------------------------
# Mixer - Settings
# -------------------------------------------------------


@dataclass
class DisplaySettings:
    gate: DisplayMode
    compressor: DisplayMode
    equaliser: DisplayMode
    equaliser_fine: DisplayMode

    def __init__(self, display: dict):
        self.gate = DisplayMode[display.get("gate")]
        self.compressor = DisplayMode[display.get("compressor")]
        self.equaliser = DisplayMode[display.get("equaliser")]
        self.equaliser_fine = DisplayMode[display.get("equaliser_fine")]


@dataclass
class MixerSettings:
    display: DisplaySettings
    mute_hold_duration: int
    vc_mute_also_mute_cm: bool

    def __init__(self, settings: dict):
        self.display = DisplaySettings(settings.get("display"))
        self.mute_hold_duration = settings.get("mute_hold_duration")
        self.vc_mute_also_mute_cm = settings.get("vc_mute_also_mute_cm")


# -------------------------------------------------------
# Mixer
# -------------------------------------------------------


@dataclass
class Mixer:
    hardware: HardwareInfo
    shutdown_commands: List[Dict[str, List[str] | str]]
    fader_status: Dict[Fader, FaderStatus]
    mic_status: MicStatus
    levels: Levels
    router: Dict[InputDevice, Dict[OutputDevice, bool]]
    cough_button: CoughButton
    lighting: Lighting
    effects: Effects
    sampler: Sampler
    settings: MixerSettings
    button_down: Dict[Button, bool]
    profile_name: str
    mic_profile_name: str

    def __init__(self, mixer: dict):
        self.hardware = HardwareInfo(mixer.get("hardware"))
        self.shutdown_commands = mixer.get("shutdown_commands")
        self.fader_status = {
            Fader[fader]: FaderStatus(status)
            for fader, status in mixer.get("fader_status").items()
        }
        self.mic_status = MicStatus(mixer.get("mic_status"))
        self.levels = Levels(mixer.get("levels"))
        self.router = {
            InputDevice[k]: {OutputDevice[k]: v for k, v in v.items()}
            for k, v in mixer.get("router").items()
        }
        self.cough_button = CoughButton(mixer.get("cough_button"))
        self.lighting = Lighting(mixer.get("lighting"))
        self.effects = Effects(mixer.get("effects"))
        self.sampler = Sampler(mixer.get("sampler"))
        self.settings = MixerSettings(mixer.get("settings"))
        self.button_down = {
            Button[button]: bool(down)
            for button, down in mixer.get("button_down").items()
        }
        self.profile_name = mixer.get("profile_name")
        self.mic_profile_name = mixer.get("mic_profile_name")


# -------------------------------------------------------
# Paths
# -------------------------------------------------------


@dataclass
class Paths:
    profiles: str
    mic_profiles: str
    samples: str
    presets: str
    icons: str
    logs: str

    def __init__(self, paths: dict):
        self.profiles = paths.get("profile_directory")
        self.mic_profiles = paths.get("mic_profile_directory")
        self.samples = paths.get("samples_directory")
        self.presets = paths.get("presets_directory")
        self.icons = paths.get("icons_directory")
        self.logs = paths.get("logs_directory")


# -------------------------------------------------------
# Files
# -------------------------------------------------------


@dataclass
class Files:
    profiles: List[str]
    mic_profiles: List[str]
    samples: Dict[str, str]
    presets: List[str]
    icons: List[str]

    def __init__(self, files: dict):
        self.profiles = files.get("profiles")
        self.mic_profiles = files.get("mic_profiles")
        self.samples = files.get("samples")
        self.presets = files.get("presets")
        self.icons = files.get("icons")


# -------------------------------------------------------
# Status
# -------------------------------------------------------


@dataclass
class Status:
    config: Config
    mixers: Dict[str, Mixer]
    paths: Paths
    files: Files

    def __init__(self, status: dict):
        self.config = Config(status.get("config"))
        self.mixers = {
            serial: Mixer(mixer) for serial, mixer in status.get("mixers").items()
        }
        self.paths = Paths(status.get("paths"))
        self.files = Files(status.get("files"))


# -------------------------------------------------------
# Patch
# -------------------------------------------------------


@dataclass
class Patch:
    operation: PatchOperation
    path: str
    value: Any

    def __init__(self, patch: dict):
        self.operation = PatchOperation[patch.get("op").title()]
        self.path = patch.get("path")
        self.value = patch.get("value")



import ctypes

from ..error import MissingFeatureError

from ..types.models import Colours
from ..types.enums import *


class GoXLRCommands:
    """
    GoXLR-specific commands.
    """

    async def __send_command(self, payload, serial=None):
        payload = {"Command": [serial or self.serial, payload]}
        return await self.send(payload)

    async def set_shutdown_commands(self, *methods) -> dict | str:
        """
        Set the commands to be executed when the GoXLR is shutting down.
        Commands are accepted as a list of lists, where the first item is
        the method name and the remaining items are the arguments for that
        method.

        :param methods: A list of methods to be executed when the GoXLR is shutting down.
        :type methods: list
        :return: The response from the GoXLR.

        :Example:

        >>> await xlr.set_shutdown_commands(
        ...     ["SetFader", Fader.A, Channel.Headphones],
        ...     ["SetFader", Fader.B, Channel.Chat]
        ... )
        """

        # Initialize the command dictionary
        command = {"SetShutdownCommands": []}

        # Iterate over the provided methods
        for method in methods:
            if not isinstance(method, list):
                raise TypeError("Methods must be provided as a list")

            # Extract the method name and arguments
            method_name, *args = method
            args = [arg.name if isinstance(arg, Enum) else arg for arg in args]

            # Create a dictionary for the method and arguments
            method_dict = {method_name: args}

            # Append the method dictionary to the command
            command["SetShutdownCommands"].append(method_dict)

        return await self.__send_command(command)

    async def set_sampler_pre_buffer_duration(self, duration: ctypes.c_uint16):
        return await self.__send_command({"SetSamplerPreBufferDuration": duration})

    async def set_fader(self, fader: Fader, channel: Channel):
        return await self.__send_command({"SetFader": [fader.name, channel.name]})

    async def set_fader_mute_function(self, fader: Fader, mute_function: MuteFunction):
        return await self.__send_command(
            {"SetFaderMuteFunction": [fader.name, mute_function.name]}
        )

    async def set_volume(self, channel: Channel, volume: ctypes.c_uint8):
        return await self.__send_command({"SetVolume": [channel.name, volume]})

    async def set_microphone_type(self, microphone_type: MicrophoneType):
        return await self.__send_command({"SetMicrophoneType": microphone_type.name})

    async def set_microphone_gain(
        self, microphone_type: MicrophoneType, gain: ctypes.c_uint16
    ):
        return await self.__send_command(
            {"SetMicrophoneGain": [microphone_type.name, gain]}
        )

    async def set_router(
        self, input_device: InputDevice, output_device: OutputDevice, enabled: bool
    ):
        return await self.__send_command(
            {"SetRouter": [input_device.name, output_device.name, enabled]}
        )

    # Cough Button
    async def set_cough_mute_function(self, mute_function: MuteFunction):
        return await self.__send_command({"SetCoughMuteFunction": mute_function.name})

    async def set_cough_is_hold(self, is_hold: bool):
        return await self.__send_command({"SetCoughIsHold": is_hold})

    # Bleep Button
    async def set_bleep_volume(self, volume: ctypes.c_int8):
        return await self.__send_command({"SetSwearButtonVolume": volume})

    # EQ Settings
    async def set_eq_mini_gain(
        self, mini_eq_frequency: MiniEqFrequency, gain: ctypes.c_int8
    ):
        return await self.__send_command(
            {"SetEqMiniGain": [mini_eq_frequency.name, gain]}
        )

    async def set_eq_mini_frequency(
        self, mini_eq_frequency: MiniEqFrequency, frequency: ctypes.c_float
    ):
        return await self.__send_command(
            {"SetEqMiniFrequency": [mini_eq_frequency.name, frequency]}
        )

    async def set_eq_gain(self, eq_frequency: EqFrequency, gain: ctypes.c_int8):
        return await self.__send_command({"SetEqGain": [eq_frequency.name, gain]})

    async def set_eq_frequency(
        self, eq_frequency: EqFrequency, frequency: ctypes.c_float
    ):
        return await self.__send_command(
            {"SetEqFrequency": [eq_frequency.name, frequency]}
        )

    # Gate Settings
    async def set_gate_threshold(self, gate_threshold: ctypes.c_int8):
        return await self.__send_command({"SetGateThreshold": gate_threshold})

    async def set_gate_attenuation(self, gate_attenuation: ctypes.c_uint8):
        return await self.__send_command({"SetGateAttenuation": gate_attenuation})

    async def set_gate_attack(self, gate_attack: GateTime):
        return await self.__send_command({"SetGateAttack": gate_attack.name})

    async def set_gate_release(self, gate_release: GateTime):
        return await self.__send_command({"SetGateRelease": gate_release.name})

    async def set_gate_active(self, gate_active: bool):
        return await self.__send_command({"SetGateActive": gate_active})

    # Compressor Settings
    async def set_compressor_threshold(self, compressor_threshold: ctypes.c_int8):
        return await self.__send_command(
            {"SetCompressorThreshold": compressor_threshold}
        )

    async def set_compressor_ratio(self, compressor_ratio: CompressorRatio):
        return await self.__send_command({"SetCompressorRatio": compressor_ratio.name})

    async def set_compressor_attack(self, compressor_attack: CompressorAttackTime):
        return await self.__send_command(
            {"SetCompressorAttack": compressor_attack.name}
        )

    async def set_compressor_release_time(
        self, compressor_release: CompressorReleaseTime
    ):
        return await self.__send_command(
            {"SetCompressorReleaseTime": compressor_release.name}
        )

    async def set_compressor_makeup_gain(self, compressor_makeup_gain: ctypes.c_int8):
        return await self.__send_command(
            {"SetCompressorMakeupGain": compressor_makeup_gain}
        )

    # Used to switch between display modes
    async def set_element_display_mode(
        self, display_mode_component: DisplayModeComponent, display_mode: DisplayMode
    ):
        return await self.__send_command(
            {"SetElementDisplayMode": [display_mode_component.name, display_mode.name]}
        )

    # DeEss
    async def set_deesser(self, deesser: ctypes.c_uint8):
        return await self.__send_command({"SetDeesser": deesser})

    # Colour Related Settings
    async def set_animation_mode(self, animation_mode: AnimationMode):
        """
        Set the animation mode of the device.

        :param animation_mode: The animation mode to set.
        :type animation_mode: AnimationMode

        :raises MissingFeatureError: If the animation mode is not supported on the GoXLR Mini.
        """
        if (
            animation_mode == AnimationMode.Ripple
            and self.get_device_type() == DeviceType.Mini
        ):
            raise MissingFeatureError("Ripple animation is not supported on the Mini")

        return await self.__send_command(
            {
                "SetAnimationMode": "None"
                if animation_mode is AnimationMode.NONE
                else animation_mode.name
            }
        )

    async def set_animation_mod1(self, animation_mod1: ctypes.c_uint8):
        return await self.__send_command({"SetAnimationMod1": animation_mod1})

    async def set_animation_mod2(self, animation_mod2: ctypes.c_uint8):
        return await self.__send_command({"SetAnimationMod2": animation_mod2})

    async def set_animation_waterfall(self, waterfall_direction: WaterfallDirection):
        return await self.__send_command(
            {"SetAnimationWaterfall": waterfall_direction.name}
        )

    async def set_global_colour(self, colour: str):
        return await self.__send_command({"SetGlobalColour": colour})

    async def set_fader_display_style(
        self, fader: Fader, fader_display_style: FaderDisplayStyle
    ):
        return await self.__send_command(
            {"SetFaderDisplayStyle": [fader.name, fader_display_style.name]}
        )

    async def set_fader_colours(self, fader: Fader, colour1: str, colour2: str):
        return await self.__send_command(
            {"SetFaderColours": [fader.name, colour1, colour2]}
        )

    async def set_all_fader_colours(self, colour1: str, colour2: str):
        return await self.__send_command({"SetAllFaderColours": [colour1, colour2]})

    async def set_all_fader_display_style(self, fader_display_style: FaderDisplayStyle):
        return await self.__send_command(
            {"SetAllFaderDisplayStyle": fader_display_style.name}
        )

    async def set_button_colours(
        self, button: Button, colour1: str, colour2: str = None
    ):
        return await self.__send_command(
            {"SetButtonColours": [button.name, colour1, colour2]}
        )

    async def set_button_off_style(
        self, button: Button, off_style: ButtonColourOffStyle
    ):
        return await self.__send_command(
            {"SetButtonOffStyle": [button.name, off_style.name]}
        )

    async def set_button_group_colours(
        self,
        button_colour_group: ButtonColourGroup,
        colour1: str,
        colour2: str = None,
    ):
        return await self.__send_command(
            {"SetButtonGroupColours": [button_colour_group.name, colour1, colour2]}
        )

    async def set_button_group_off_style(
        self, button_colour_group: ButtonColourGroup, off_style: ButtonColourOffStyle
    ):
        return await self.__send_command(
            {"SetButtonGroupOffStyle": [button_colour_group.name, off_style.name]}
        )

    async def set_simple_colour(
        self, simple_colour_target: SimpleColourTarget, colour: str
    ):
        return await self.__send_command(
            {"SetSimpleColour": [simple_colour_target.name, colour]}
        )

    async def set_encoder_colours(self, encoder: EncoderColourTarget, colours: Colours):
        colour1 = colours.colour_one
        colour2 = colours.colour_two
        colour3 = colours.colour_three
        return await self.__send_command(
            {"SetEncoderColour": [encoder.name, colour1, colour2, colour3]}
        )

    async def set_sample_colours(self, sample: SamplerColourTarget, colours: Colours):
        colour1 = colours.colour_one
        colour2 = colours.colour_two
        colour3 = colours.colour_three
        return await self.__send_command(
            {"SetSampleColour": [sample.name, colour1, colour2, colour3]}
        )

    async def set_sample_off_style(
        self, sample: SamplerColourTarget, off_style: ButtonColourOffStyle
    ):
        return await self.__send_command(
            {"SetSampleOffStyle": [sample.name, off_style.name]}
        )

    # Effect Related Settings
    async def load_effect(self, effect_name: str):
        return await self.__send_command({"LoadEffect": effect_name})

    async def rename_active_preset(self, new_name: str):
        return await self.__send_command({"RenameActivePreset": new_name})

    async def save_active_preset(self):
        return await self.__send_command({"SaveActivePreset": []})

    # Reverb
    async def set_reverb_style(self, reverb_style: ReverbStyle):
        return await self.__send_command({"SetReverbStyle": reverb_style.name})

    async def set_reverb_amount(self, reverb_amount: ctypes.c_uint8):
        return await self.__send_command({"SetReverbAmount": reverb_amount})

    async def set_reverb_decay(self, reverb_decay: ctypes.c_uint16):
        return await self.__send_command({"SetReverbDecay": reverb_decay})

    async def set_reverb_early_level(self, reverb_early_level: ctypes.c_int8):
        return await self.__send_command({"SetReverbEarlyLevel": reverb_early_level})

    async def set_reverb_tail_level(self, reverb_tail_level: ctypes.c_int8):
        return await self.__send_command({"SetReverbTailLevel": reverb_tail_level})

    async def set_reverb_pre_delay(self, reverb_pre_delay: ctypes.c_uint8):
        return await self.__send_command({"SetReverbPreDelay": reverb_pre_delay})

    async def set_reverb_low_colour(self, reverb_low_colour: ctypes.c_int8):
        return await self.__send_command({"SetReverbLowColour": reverb_low_colour})

    async def set_reverb_high_colour(self, reverb_high_colour: ctypes.c_int8):
        return await self.__send_command({"SetReverbHighColour": reverb_high_colour})

    async def set_reverb_high_factor(self, reverb_high_factor: ctypes.c_int8):
        return await self.__send_command({"SetReverbHighFactor": reverb_high_factor})

    async def set_reverb_diffuse(self, reverb_diffuse: ctypes.c_uint8):
        return await self.__send_command({"SetReverbDiffuse": reverb_diffuse})

    async def set_reverb_mod_speed(self, reverb_mod_speed: ctypes.c_uint8):
        return await self.__send_command({"SetReverbModSpeed": reverb_mod_speed})

    async def set_reverb_mod_depth(self, reverb_mod_depth: ctypes.c_uint8):
        return await self.__send_command({"SetReverbModDepth": reverb_mod_depth})

    # Echo
    async def set_echo_style(self, echo_style: EchoStyle):
        return await self.__send_command({"SetEchoStyle": echo_style.name})

    async def set_echo_amount(self, echo_amount: ctypes.c_uint8):
        return await self.__send_command({"SetEchoAmount": echo_amount})

    async def set_echo_feedback(self, echo_feedback: ctypes.c_uint8):
        return await self.__send_command({"SetEchoFeedback": echo_feedback})

    async def set_echo_tempo(self, echo_tempo: ctypes.c_uint16):
        return await self.__send_command({"SetEchoTempo": echo_tempo})

    async def set_echo_delay_left(self, echo_delay_left: ctypes.c_uint16):
        return await self.__send_command({"SetEchoDelayLeft": echo_delay_left})

    async def set_echo_delay_right(self, echo_delay_right: ctypes.c_uint16):
        return await self.__send_command({"SetEchoDelayRight": echo_delay_right})

    async def set_echo_feedback_left(self, echo_feedback_left: ctypes.c_uint8):
        return await self.__send_command({"SetEchoFeedbackLeft": echo_feedback_left})

    async def set_echo_feedback_right(self, echo_feedback_right: ctypes.c_uint8):
        return await self.__send_command({"SetEchoFeedbackRight": echo_feedback_right})

    async def set_echo_feedback_xfb_l_to_r(
        self, echo_feedback_xfb_l_to_r: ctypes.c_uint8
    ):
        return await self.__send_command(
            {"SetEchoFeedbackXFBLtoR": echo_feedback_xfb_l_to_r}
        )

    async def set_echo_feedback_xfb_r_to_l(
        self, echo_feedback_xfb_r_to_l: ctypes.c_uint8
    ):
        return await self.__send_command(
            {"SetEchoFeedbackXFBRtoL": echo_feedback_xfb_r_to_l}
        )

    # Pitch
    async def set_pitch_style(self, pitch_style: PitchStyle):
        return await self.__send_command({"SetPitchStyle": pitch_style.name})

    async def set_pitch_amount(self, pitch_amount: ctypes.c_int8):
        return await self.__send_command({"SetPitchAmount": pitch_amount})

    async def set_pitch_character(self, pitch_character: ctypes.c_uint8):
        return await self.__send_command({"SetPitchCharacter": pitch_character})

    # Gender
    async def set_gender_style(self, gender_style: GenderStyle):
        return await self.__send_command({"SetGenderStyle": gender_style.name})

    async def set_gender_amount(self, gender_amount: ctypes.c_int8):
        return await self.__send_command({"SetGenderAmount": gender_amount})

    # Megaphone
    async def set_megaphone_style(self, megaphone_style: MegaphoneStyle):
        return await self.__send_command({"SetMegaphoneStyle": megaphone_style.name})

    async def set_megaphone_amount(self, megaphone_amount: ctypes.c_uint8):
        return await self.__send_command({"SetMegaphoneAmount": megaphone_amount})

    async def set_megaphone_post_gain(self, megaphone_post_gain: ctypes.c_int8):
        return await self.__send_command({"SetMegaphonePostGain": megaphone_post_gain})

    # Robot
    async def set_robot_style(self, robot_style: RobotStyle):
        return await self.__send_command({"SetRobotStyle": robot_style.name})

    async def set_robot_gain(self, robot_range: RobotRange, robot_gain: ctypes.c_int8):
        return await self.__send_command(
            {"SetRobotGain": [robot_range.name, robot_gain]}
        )

    async def set_robot_frequency(
        self, robot_range: RobotRange, robot_frequency: ctypes.c_uint8
    ):
        return await self.__send_command(
            {"SetRobotFreq": [robot_range.name, robot_frequency]}
        )

    async def set_robot_width(
        self, robot_range: RobotRange, robot_width: ctypes.c_uint8
    ):
        return await self.__send_command(
            {"SetRobotWidth": [robot_range.name, robot_width]}
        )

    async def set_robot_waveform(self, robot_waveform: ctypes.c_uint8):
        return await self.__send_command({"SetRobotWaveform": robot_waveform})

    async def set_robot_pulse_width(self, robot_pulse_width: ctypes.c_uint8):
        return await self.__send_command({"SetRobotPulseWidth": robot_pulse_width})

    async def set_robot_threshold(self, robot_threshold: ctypes.c_int8):
        return await self.__send_command({"SetRobotThreshold": robot_threshold})

    async def set_robot_dry_mix(self, robot_dry_mix: ctypes.c_int8):
        return await self.__send_command({"SetRobotDryMix": robot_dry_mix})

    # Hardtune
    async def set_hardtune_style(self, hardtune_style: HardTuneStyle):
        return await self.__send_command({"SetHardTuneStyle": hardtune_style.name})

    async def set_hardtune_amount(self, hardtune_amount: ctypes.c_uint8):
        return await self.__send_command({"SetHardTuneAmount": hardtune_amount})

    async def set_hardtune_rate(self, hardtune_rate: ctypes.c_uint8):
        return await self.__send_command({"SetHardTuneRate": hardtune_rate})

    async def set_hardtune_window(self, hardtune_window: ctypes.c_uint16):
        return await self.__send_command({"SetHardTuneWindow": hardtune_window})

    async def set_hardtune_source(self, hardtune_source: HardTuneSource):
        return await self.__send_command({"SetHardTuneSource": hardtune_source.name})

    # Sampler
    async def clear_sample_process_error(self):
        return await self.__send_command({"ClearSampleProcessError": []})

    async def set_sampler_function(
        self,
        sample_bank: SampleBank,
        sample_button: SampleButton,
        sample_playback_mode: SamplePlaybackMode,
    ):
        return await self.__send_command(
            {
                "SetSamplerFunction": [
                    sample_bank.name,
                    sample_button.name,
                    sample_playback_mode.name,
                ]
            }
        )

    async def set_sampler_order(
        self,
        sample_bank: SampleBank,
        sample_button: SampleButton,
        sample_play_order: SamplePlayOrder,
    ):
        return await self.__send_command(
            {
                "SetSamplerOrder": [
                    sample_bank.name,
                    sample_button.name,
                    sample_play_order.name,
                ]
            }
        )

    async def add_sample(
        self, sample_bank: SampleBank, sample_button: SampleButton, sample_name: str
    ):
        return await self.__send_command(
            {"AddSample": [sample_bank.name, sample_button.name, sample_name]}
        )

    async def set_sample_start_percent(
        self,
        sample_bank: SampleBank,
        sample_button: SampleButton,
        index: int,
        sample_start_percent: ctypes.c_float,
    ):
        return await self.__send_command(
            {
                "SetSampleStartPercent": [
                    sample_bank.name,
                    sample_button.name,
                    index,
                    sample_start_percent,
                ]
            }
        )

    async def set_sample_stop_percent(
        self,
        sample_bank: SampleBank,
        sample_button: SampleButton,
        index: int,
        sample_stop_percent: ctypes.c_float,
    ):
        return await self.__send_command(
            {
                "SetSampleStopPercent": [
                    sample_bank.name,
                    sample_button.name,
                    index,
                    sample_stop_percent,
                ]
            }
        )

    async def remove_sample_by_index(
        self, sample_bank: SampleBank, sample_button: SampleButton, index: int
    ):
        return await self.__send_command(
            {"RemoveSampleByIndex": [sample_bank.name, sample_button.name, index]}
        )

    async def play_sample_by_index(
        self, sample_bank: SampleBank, sample_button: SampleButton, index: int
    ):
        return await self.__send_command(
            {"PlaySampleByIndex": [sample_bank.name, sample_button.name, index]}
        )

    async def play_next_sample(
        self, sample_bank: SampleBank, sample_button: SampleButton
    ):
        return await self.__send_command(
            {"PlayNextSample": [sample_bank.name, sample_button.name]}
        )

    async def stop_sample_playback(
        self, sample_bank: SampleBank, sample_button: SampleButton
    ):
        return await self.__send_command(
            {"StopSamplePlayback": [sample_bank.name, sample_button.name]}
        )

    # Scribbles
    async def set_scribble_icon(self, fader: Fader, scribble_icon: str = None):
        return await self.__send_command(
            {"SetScribbleIcon": [fader.name, scribble_icon]}
        )

    async def set_scribble_text(self, fader: Fader, scribble_text: str):
        return await self.__send_command(
            {"SetScribbleText": [fader.name, scribble_text]}
        )

    async def set_scribble_number(self, fader: Fader, scribble_number: str):
        return await self.__send_command(
            {"SetScribbleNumber": [fader.name, scribble_number]}
        )

    async def set_scribble_invert(self, fader: Fader, scribble_invert: bool):
        return await self.__send_command(
            {"SetScribbleInvert": [fader.name, scribble_invert]}
        )

    # Profile Handling
    async def new_profile(self, profile_name: str):
        return await self.__send_command({"NewProfile": profile_name})

    async def load_profile(self, profile_name: str, save_changes: bool = False):
        return await self.__send_command({"LoadProfile": [profile_name, save_changes]})

    async def load_profile_colours(self, profile_name: str):
        return await self.__send_command({"LoadProfileColours": profile_name})

    async def save_profile(self):
        return await self.__send_command({"SaveProfile": []})

    async def save_profile_as(self, profile_name: str):
        return await self.__send_command({"SaveProfileAs": profile_name})

    async def delete_profile(self, profile_name: str):
        return await self.__send_command({"DeleteProfile": profile_name})

    async def new_mic_profile(self, profile_name: str):
        return await self.__send_command({"NewMicProfile": profile_name})

    async def load_mic_profile(self, profile_name: str, save_changes: bool = False):
        return await self.__send_command(
            {"LoadMicProfile": [profile_name, save_changes]}
        )

    async def save_mic_profile(self):
        return await self.__send_command({"SaveMicProfile": []})

    async def save_mic_profile_as(self, profile_name: str):
        return await self.__send_command({"SaveMicProfileAs": profile_name})

    async def delete_mic_profile(self, profile_name: str):
        return await self.__send_command({"DeleteMicProfile": profile_name})

    # General Settings
    async def set_mute_hold_duration(self, mute_hold_duration: ctypes.c_uint16):
        return await self.__send_command({"SetMuteHoldDuration": mute_hold_duration})

    async def set_vc_mute_also_mute_cm(self, vc_mute_also_mute_cm: bool):
        return await self.__send_command({"SetVCMuteAlsoMuteCM": vc_mute_also_mute_cm})

    # These control the current GoXLR state
    async def set_active_effect_preset(self, active_effect_preset: EffectBankPreset):
        return await self.__send_command(
            {"SetActiveEffectPreset": active_effect_preset.name}
        )

    async def set_sampler_active_bank(self, active_sampler_bank: SampleBank):
        return await self.__send_command(
            {"SetActiveSamplerBank": active_sampler_bank.name}
        )

    async def set_megaphone_enabled(self, megaphone_enabled: bool):
        return await self.__send_command({"SetMegaphoneEnabled": megaphone_enabled})

    async def set_robot_enabled(self, robot_enabled: bool):
        return await self.__send_command({"SetRobotEnabled": robot_enabled})

    async def set_hardtune_enabled(self, hardtune_enabled: bool):
        return await self.__send_command({"SetHardTuneEnabled": hardtune_enabled})

    async def set_fx_enabled(self, fx_enabled: bool):
        return await self.__send_command({"SetFXEnabled": fx_enabled})

    async def set_fader_mute_state(self, fader: Fader, mute_state: MuteState):
        return await self.__send_command(
            {"SetFaderMuteState": [fader.name, mute_state.name]}
        )

    async def set_cough_mute_state(self, mute_state: MuteState):
        return await self.__send_command({"SetCoughMuteState": mute_state.name})

    # Submix commands
    async def set_submix_enabled(self, enabled: bool):
        return await self.__send_command({"SetSubMixEnabled": enabled})

    async def set_submix_volume(self, channel: SubMixChannel, volume: ctypes.c_uint8):
        return await self.__send_command({"SetSubMixVolume": [channel.name, volume]})

    async def set_submix_linked(self, channel: SubMixChannel, linked: bool):
        return await self.__send_command({"SetSubMixLinked": [channel.name, linked]})

    async def set_submix_output_mix(self, output_device: OutputDevice, mix: Mix):
        return await self.__send_command(
            {"SetSubMixOutputMix": [output_device.name, mix.name]}
        )

    # Mix monitoring
    async def set_monitor_mix(self, output_device: OutputDevice):
        return await self.__send_command({"SetMonitorMix": output_device.name})

    # it's..done..?



from typing import Union
from ..error import DaemonError, MixerNotFoundError
from ..types.models import *


class StatusCommands:
    """
    Used to retrieve information from the GoXLR Utility daemon.
    """

    async def get_status(self) -> Status:
        """
        Returns the status of the GoXLR.

        :return: The status of the GoXLR.

        :raises DaemonError: If the status could not be retrieved.

        :note: You should use GoXLR.update() instead of this method.
        """
        status = await self.send("GetStatus")

        if not status:
            raise DaemonError("Failed to get status from daemon.")

        return Status(status)

    # ------------------------------------------------------------
    # Config
    # ------------------------------------------------------------

    def get_http_settings(self) -> HttpSettings:
        """
        :return: The HTTP settings of the GoXLR.
        """
        return self.status.config.http_settings

    def get_daemon_version(self) -> str:
        """
        :return: The version of the GoXLR Utility daemon.
        """
        return self.status.config.daemon_version

    def is_autostart_enabled(self) -> bool:
        """
        :return: Whether or not the GoXLR Utility daemon is set to start on boot.
        """
        return self.status.config.autostart_enabled

    def is_tray_icon_visible(self) -> bool:
        """
        :return: Whether or not the GoXLR Utility daemon is set to show a tray icon.
        """
        return self.status.config.show_tray_icon

    def is_tts_enabled(self) -> bool:
        """
        :return: Whether or not TTS is enabled.
        """
        return self.status.config.tts_enabled

    def is_network_access_allowed(self) -> bool:
        """
        :return: Whether or not network access is allowed.
        """
        return self.status.config.allow_network_access

    def get_log_level(self) -> LogLevel:
        """
        :return: The log level of the GoXLR Utility daemon.
        """
        return self.status.config.log_level

    # ------------------------------------------------------------
    # Mixers
    # ------------------------------------------------------------

    def get_mixer(self, serial: str = None) -> Mixer:
        """
        Returns a mixer object with the specified serial number.

        :param serial: The serial number of the mixer to interact with.
                          If not specified, it will default to the currently selected mixer.

        :return: The requested mixer object.

        :raises MixerNotFoundError: If the specified mixer is not found.
        """
        if not serial:
            return self.mixer  # return the currently selected mixer

        if serial not in self.status.mixers:
            raise MixerNotFoundError(f"Mixer {serial} not found.")

        return self.status.mixers.get(serial)

    # assume that the user wants to interact with the currently selected mixer
    # using self.mixer

    # Hardware Info

    def get_hardware_info(self) -> HardwareInfo:
        return self.mixer.hardware

    def get_versions(self) -> MixerVersions:
        return self.get_hardware_info().versions

    def get_serial_number(self) -> str:
        return self.get_hardware_info().serial_number

    def get_manufactured_date(self) -> datetime:
        return self.get_hardware_info().manufactured_date

    def get_device_type(self) -> DeviceType:
        return self.get_hardware_info().device_type

    def get_usb_device(self) -> USBDevice:
        return self.get_hardware_info().usb_device

    # Shutdown Commands

    def get_shutdown_commands(self) -> List[Dict[str, List[str] | str]]:
        return self.mixer.shutdown_commands

    # Fader Status

    def get_fader_status(self) -> Dict[Fader, FaderStatus]:
        return self.mixer.fader_status

    def get_fader(self, fader: Fader) -> FaderStatus | None:
        return self.get_fader_status().get(fader)

    def get_fader_channel(self, fader: Fader) -> Channel:
        return self.get_fader(fader).channel

    def get_fader_mute_function(self, fader: Fader) -> MuteFunction:
        return self.get_fader(fader).mute_type

    def get_fader_scribble(self, fader: Fader) -> Scribble:
        return self.get_fader(fader).scribble

    def get_fader_scribble_file_name(self, fader: Fader) -> str:
        return self.get_fader_scribble(fader).file_name

    def get_scribble_text(self, fader: Fader) -> str:
        return self.get_fader_scribble(fader).bottom_text

    def get_scribble_left_text(self, fader: Fader) -> str:
        return self.get_fader_scribble(fader).left_text

    def get_scribble_invert(self, fader: Fader) -> bool:
        return self.get_fader_scribble(fader).inverted

    def get_fader_mute_state(self, fader: Fader) -> MuteState:
        return self.get_fader(fader).mute_state

    def is_fader_muted(self, fader: Fader) -> bool:
        """
        Helper method to check if a fader is muted.
        """
        return self.get_fader_mute_state(fader) != MuteState.Unmuted

    # Mic Status

    def get_mic_status(self) -> MicStatus:
        return self.mixer.mic_status

    def get_microphone_type(self) -> MicrophoneType:
        return self.get_mic_status().mic_type

    def get_microphone_gains(self) -> Dict[MicrophoneType, int]:
        return self.get_mic_status().mic_gains

    def get_microphone_gain(self, mic_type: MicrophoneType = None) -> int | None:
        """
        :note: If no mic type is specified, the currently selected mic type will be used.
        """
        if not mic_type:
            mic_type = self.get_mic_type()
        return self.get_mic_gains().get(mic_type)

    def get_eq(self) -> Equaliser:
        return self.get_mic_status().equaliser

    def get_eq_gain(self, frequency: EqFrequency) -> int | None:
        return self.get_eq().gain.get(frequency)

    def get_eq_frequency(self, frequency: EqFrequency) -> float | None:
        return self.get_eq().frequency.get(frequency)

    def get_eq_mini(self) -> EqMini:
        return self.get_mic_status().equaliser_mini

    def get_eq_mini_gain(self, frequency: MiniEqFrequency) -> int | None:
        return self.get_eq_mini().gain.get(frequency)

    def get_eq_mini_frequency(self, frequency: MiniEqFrequency) -> float | None:
        return self.get_eq_mini().frequency.get(frequency)

    def get_noise_gate(self) -> NoiseGate:
        return self.get_mic_status().noise_gate

    def get_compressor(self) -> Compressor:
        return self.get_mic_status().compressor

    # Levels

    def get_levels(self) -> Levels:
        return self.mixer.levels

    def is_submix_supported(self) -> bool:
        return self.get_levels().submix_supported

    def get_monitor_mix(self) -> OutputDevice:
        return self.get_levels().output_monitor

    def get_volumes(self) -> Dict[Channel, int]:
        return self.get_levels().volumes

    def get_volume(self, channel: Channel) -> int | None:
        return self.get_volumes().get(channel)

    def get_submixes(self) -> Submixes | None:
        return self.get_levels().submix

    def get_submix_input_mix(self, channel: SubMixChannel) -> Submix | None:
        return self.get_submixes().inputs.get(channel)

    def get_submix_volume(self, channel: SubMixChannel) -> int | None:
        return self.get_submix_input_mix(channel).volume

    def get_submix_linked(self, channel: SubMixChannel) -> bool | None:
        return self.get_submix_input_mix(channel).linked

    def get_submix_output_mix(self, output: OutputDevice) -> Mix | None:
        return self.get_submixes().outputs.get(output)

    def get_bleep_volume(self) -> int:
        return self.get_levels().bleep

    def get_deesser(self) -> int:
        return self.get_levels().deess

    # Router

    def get_routing_table(self) -> Dict[InputDevice, Dict[OutputDevice, bool]]:
        return self.mixer.router

    def get_routed_outputs(self, input: InputDevice) -> Dict[OutputDevice, bool] | None:
        return self.get_routing_table().get(input)

    def get_router(self, input: InputDevice, output: OutputDevice) -> bool | None:
        # Would rather name this is_routed, but it's to comply with set_router
        return self.get_routed_outputs(input).get(output)

    def get_routed_inputs(self, output: OutputDevice) -> List[InputDevice]:
        inputs = []
        for input, outputs in self.get_routing_table().items():
            if outputs.get(output):
                inputs.append(input)
        return inputs

    # Cough button

    def get_cough_button(self) -> CoughButton:
        return self.mixer.cough_button

    def get_cough_is_hold(self) -> bool:
        return not self.get_cough_button().is_toggle

    def get_cough_mute_function(self) -> MuteFunction:
        return self.get_cough_button().mute_type

    def get_cough_mute_state(self) -> MuteState:
        return self.get_cough_button().mute_state

    def is_cough_button_muted(self) -> bool:
        return self.get_cough_mute_state() != MuteState.Unmuted

    # Lighting

    def get_lighting(self) -> Lighting:
        return self.mixer.lighting

    def get_animation(self) -> Animation:
        return self.get_lighting().animation

    def is_animation_supported(self) -> bool:
        return self.get_animation().supported

    def get_animation_mode(self) -> AnimationMode:
        return self.get_animation().mode

    def get_animation_mod1(self) -> int:
        return self.get_animation().mod1

    def get_animation_mod2(self) -> int:
        return self.get_animation().mod2

    def get_animation_waterfall(self) -> WaterfallDirection:
        return self.get_animation().waterfall_direction

    def get_fader_lighting(self, fader: Fader) -> FaderLighting | None:
        return self.get_lighting().faders.get(fader)

    def get_fader_display_style(self, fader: Fader) -> FaderDisplayStyle | None:
        return self.get_fader_lighting().style

    def get_fader_colours(self, fader: Fader) -> Colours | None:
        return self.get_fader_lighting().colours

    def get_button_lighting(self, button: Button) -> ButtonLighting | None:
        return self.get_lighting().buttons.get(button)

    def get_button_colours(self, button: Button) -> Colours | None:
        return self.get_button_lighting(button).colours

    def get_button_off_style(self, button: Button) -> ButtonColourOffStyle | None:
        return self.get_button_lighting(button).off_style

    def get_simple_colour(self, target: SimpleColourTarget) -> str | None:
        return self.get_lighting().simple.get(target).colour_one

    def get_sample_lighting(self, target: SamplerColourTarget) -> ButtonLighting | None:
        return self.get_lighting().sampler.get(target)

    def get_sample_colours(self, target: SamplerColourTarget) -> Colours | None:
        return self.get_sample_lighting(target).colours

    def get_encoder_colours(self, encoder: Encoder) -> Colours | None:
        return self.get_lighting().encoders.get(encoder).colours

    # Effects

    def get_effects(self) -> Effects:
        return self.mixer.effects

    def is_effects_enabled(self) -> bool:
        return self.get_effects().is_enabled

    def get_active_effect_preset(self) -> EffectBankPreset:
        return self.get_effects().active_preset

    def get_preset_name(self, preset: EffectBankPreset) -> str | None:
        return self.get_effects().preset_names.get(preset)

    def get_current_effects(self) -> CurrentEffects:
        return self.get_effects().current

    def get_reverb(self) -> Reverb:
        return self.get_current_effects().reverb

    def get_reverb_amount(self) -> int:
        return self.get_reverb().amount

    def get_reverb_decay(self) -> int:
        return self.get_reverb().decay

    def get_reverb_diffuse(self) -> int:
        return self.get_reverb().diffuse

    def get_reverb_early_level(self) -> int:
        return self.get_reverb().early_level

    def get_reverb_high_colour(self) -> int:
        return self.get_reverb().hi_colour

    def get_reverb_high_factor(self) -> int:
        return self.get_reverb().hi_factor

    def get_reverb_low_colour(self) -> int:
        return self.get_reverb().lo_colour

    def get_reverb_mod_depth(self) -> int:
        return self.get_reverb().mod_depth

    def get_reverb_mod_speed(self) -> int:
        return self.get_reverb().mod_speed

    def get_reverb_pre_delay(self) -> int:
        return self.get_reverb().pre_delay

    def get_reverb_style(self) -> ReverbStyle:
        return self.get_reverb().style

    def get_reverb_tail_level(self) -> int:
        return self.get_reverb().tail_level

    def get_echo(self) -> Echo:
        return self.get_current_effects().echo

    def get_echo_amount(self) -> int:
        return self.get_echo().amount

    def get_echo_delay_left(self) -> int:
        return self.get_echo().delay_left

    def get_echo_delay_right(self) -> int:
        return self.get_echo().delay_right

    def get_echo_feedback(self) -> int:
        return self.get_echo().feedback

    def get_echo_feedback_left(self) -> int:
        return self.get_echo().feedback_left

    def get_echo_feedback_right(self) -> int:
        return self.get_echo().feedback_right

    def get_echo_feedback_xfb_l_to_r(self) -> int:
        return self.get_echo().feedback_xfb_l_to_r

    def get_echo_feedback_xfb_r_to_l(self) -> int:
        return self.get_echo().feedback_xfb_r_to_l

    def get_echo_style(self) -> EchoStyle:
        return self.get_echo().style

    def get_echo_tempo(self) -> int:
        return self.get_echo().tempo

    def get_pitch(self) -> Pitch:
        return self.get_current_effects().pitch

    def get_pitch_amount(self) -> int:
        return self.get_pitch().amount

    def get_pitch_character(self) -> int:
        return self.get_pitch().character

    def get_pitch_style(self) -> PitchStyle:
        return self.get_pitch().style

    def get_gender(self) -> Gender:
        return self.get_current_effects().gender

    def get_gender_amount(self) -> int:
        return self.get_gender().amount

    def get_gender_style(self) -> GenderStyle:
        return self.get_gender().style

    def get_megaphone(self) -> Megaphone:
        return self.get_current_effects().megaphone

    def get_megaphone_amount(self) -> int:
        return self.get_megaphone().amount

    def is_megaphone_enabled(self) -> bool:
        return self.get_megaphone().is_enabled

    def get_megaphone_post_gain(self) -> int:
        return self.get_megaphone().post_gain

    def get_megaphone_style(self) -> MegaphoneStyle:
        return self.get_megaphone().style

    def get_robot(self) -> Robot:
        return self.get_current_effects().robot

    def get_robot_dry_mix(self) -> int:
        return self.get_robot().dry_mix

    def is_robot_enabled(self) -> bool:
        return self.get_robot().is_enabled

    def get_robot_frequency(self, range: RobotRange) -> int:
        match range:
            case RobotRange.Low:
                return self.get_robot().low_freq
            case RobotRange.Medium:
                return self.get_robot().mid_freq
            case RobotRange.High:
                return self.get_robot().high_freq

    def get_robot_gain(self, range: RobotRange) -> int:
        match range:
            case RobotRange.Low:
                return self.get_robot().low_gain
            case RobotRange.Medium:
                return self.get_robot().mid_gain
            case RobotRange.High:
                return self.get_robot().high_gain

    def get_robot_width(self, range: RobotRange) -> int:
        match range:
            case RobotRange.Low:
                return self.get_robot().low_width
            case RobotRange.Medium:
                return self.get_robot().mid_width
            case RobotRange.High:
                return self.get_robot().high_width

    def get_robot_pulse_width(self) -> int:
        return self.get_robot().pulse_width

    def get_robot_style(self) -> RobotStyle:
        return self.get_robot().style

    def get_robot_threshold(self) -> int:
        return self.get_robot().threshold

    def get_robot_waveform(self) -> int:
        return self.get_robot().waveform

    def get_hardtune(self) -> HardTune:
        return self.get_current_effects().hard_tune

    def is_hardtune_enabled(self) -> bool:
        return self.get_hardtune().is_enabled

    def get_hardtune_amount(self) -> int:
        return self.get_hardtune().amount

    def get_hardtune_rate(self) -> int:
        return self.get_hardtune().rate

    def get_hardtune_source(self) -> HardTuneSource:
        return self.get_hardtune().source

    def get_hardtune_style(self) -> HardTuneStyle:
        return self.get_hardtune().style

    def get_hardtune_window(self) -> int:
        return self.get_hardtune().window

    # Sampler

    def get_sampler(self) -> Sampler:
        return self.mixer.sampler

    def get_sampler_processing_state(self) -> SamplerProcessState:
        return self.get_sampler().processing_state

    def get_sampler_active_bank(self) -> SampleBank:
        return self.get_sampler().active_bank

    def is_clear_sample_process_error(self) -> bool:
        # Not entirely sure what this does?
        return self.get_sampler().clear_active

    def get_sampler_pre_buffer_duration(self) -> int:
        return self.get_sampler().record_buffer

    def get_sample_banks(self) -> Dict[SampleBank, Dict[SampleButton, SampleMetadata]]:
        return self.get_sampler().banks

    def get_sample_bank(
        self, bank: SampleBank
    ) -> Dict[SampleButton, SampleMetadata] | None:
        return self.get_sample_banks().get(bank)

    def get_sample_metadata(
        self, bank: SampleBank, button: SampleButton
    ) -> SampleMetadata | None:
        return self.get_sample_bank(bank).get(button)

    def get_sampler_function(
        self, bank: SampleBank, button: SampleButton
    ) -> SamplePlaybackMode | None:
        return self.get_sample_metadata(bank, button).function

    def get_sampler_order(
        self, bank: SampleBank, button: SampleButton
    ) -> SamplePlayOrder | None:
        return self.get_sample_metadata(bank, button).order

    def get_samples(
        self, bank: SampleBank, button: SampleButton
    ) -> List[Sample] | None:
        return self.get_sample_metadata(bank, button).samples

    def get_sample_start_percent(
        self, bank: SampleBank, button: SampleButton, index: int
    ) -> float | None:
        return self.get_samples(bank, button)[index].start_pct

    def get_sample_stop_percent(
        self, bank: SampleBank, button: SampleButton, index: int
    ) -> float | None:
        return self.get_samples(bank, button)[index].stop_pct

    def get_sample_is_playing(self, bank: SampleBank, button: SampleButton) -> bool:
        return self.get_sample_metadata(bank, button).is_playing

    def get_sample_is_recording(self, bank: SampleBank, button: SampleButton) -> bool:
        return self.get_sample_metadata(bank, button).is_recording

    # Settings

    def get_settings(self) -> MixerSettings:
        return self.mixer.settings

    def get_display(self) -> DisplaySettings:
        return self.get_settings().display

    def get_noise_gate_display_mode(self) -> DisplayMode:
        return self.get_display().gate

    def get_compressor_display_mode(self) -> DisplayMode:
        return self.get_display().compressor

    def get_equaliser_display_mode(self) -> DisplayMode:
        return self.get_display().equaliser

    def get_equaliser_fine_display_mode(self) -> DisplayMode:
        return self.get_display().equaliser_fine

    def get_mute_hold_duration(self) -> int:
        return self.get_settings().mute_hold_duration

    def is_vc_mute_also_mute_cm(self) -> bool:
        return self.get_settings().vc_mute_also_mute_cm

    # Button down

    def get_button_down(self, button: Button) -> bool:
        return self.mixer.button_down[button]

    async def wait_for_button(
        self,
        buttons: Union[List[Button], Dict[Button, bool]],
        all_values: bool = False,
        invert: bool = False,
    ) -> Dict[Button, bool]:
        """
        Waits for the specified button states to be achieved.

        :param buttons: List of buttons or dictionary specifying the button states to wait for.
                        - If a list, all buttons are assumed to have the desired state as True (down).
                        - If a dictionary, Key: Button, Value: Desired button state (True for down, False for up).
        :param all: Whether to wait for all (instead of any) of the buttons to achieve the desired state.
        :param invert: Whether to check for the opposite button states.
        """
        if isinstance(buttons, dict):
            button_states = buttons
        else:
            button_states = {button: True for button in buttons}

        if invert:
            button_states = {
                button: not state for button, state in button_states.items()
            }

        if all_values:
            while not all(
                self.get_button_down(button) == state
                for button, state in button_states.items()
            ):
                await self.receive_patch()
        else:
            while not any(
                self.get_button_down(button) == state
                for button, state in button_states.items()
            ):
                await self.receive_patch()

        await self.update()  # be sure it's up to date

        return {button: self.get_button_down(button) for button in button_states.keys()}

    # Profile name

    def get_profile_name(self) -> str:
        return self.mixer.profile_name

    def get_mic_profile_name(self) -> str:
        return self.mixer.mic_profile_name

    # ------------------------------------------------------------
    # Paths
    # ------------------------------------------------------------

    def get_path(self, path: PathType) -> str:
        """
        Returns the path of the specified path type.

        :param path: The path type to get the path of.

        :return: The path of the specified path type.

        :raises ValueError: If the specified path type is invalid.
        """
        match path:
            case PathType.Profiles:
                return self.status.files.profiles
            case PathType.MicProfiles:
                return self.status.files.mic_profiles
            case PathType.Samples:
                return self.status.files.samples
            case PathType.Presets:
                return self.status.files.presets
            case PathType.Icons:
                return self.status.files.icons
            case PathType.Logs:
                return self.status.files.logs
            case _:
                raise ValueError(f"Invalid path type: {path}")

    # ------------------------------------------------------------
    # Files
    # ------------------------------------------------------------

    def get_files(self, path: PathType) -> list:
        """
        Returns the files of the specified path type.

        :param path: The path type to get the files of.

        :return: The files of the specified path type.

        :raises ValueError: If the specified path type is invalid.
        """
        match path:
            case PathType.Profiles:
                return self.status.files.profiles_files
            case PathType.MicProfiles:
                return self.status.files.mic_profiles
            case PathType.Samples:
                return self.status.files.samples
            case PathType.Presets:
                return self.status.files.presets
            case PathType.Icons:
                return self.status.files.icons
            case _:
                raise ValueError(f"Invalid path type: {path}")



from setuptools import setup, find_packages

# Thanks to https://stackoverflow.com/a/7071358
import re

VERSIONFILE = "goxlr/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
    name="goxlr",
    version=verstr,
    description="A Python wrapper for the GoXLR Utility API.",
    url="https://github.com/samcarsonx/goxlr-py",
    author="Sam Carson",
    author_email="sam@samcarson.co.uk",
    license="MIT",
    packages=find_packages(),
    install_requires=["asyncio", "websockets"],
    python_requires=">=3.10",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
)
